//
// Copyright 1997 Steve Passe
//
// Permission to use, copy, modify, and distribute this software and its
// documentation for any purpose and without fee is hereby granted, provided
// that the above copyright notice appear in all copies and that both that
// copyright notice and this permission notice appear in supporting
// documentation, and that the name of AUTHORS (Steve Passe and Paul Crook)
// are not be used in advertising or publicity pertaining to distribution of
// the software without specific, written prior permission.  AUTHORS make no
// representations about the suitability of this software for any purpose.
// It is provided "as is" without express or implied warranty.
//
// AUTHORS DISCLAIM ALL WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING ALL
// IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO EVENT SHALL AUTHORS
// BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
// WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION
// OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN 
// CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
//
// 1997 Revision
// AUTHOR:     Steve Passe <smp@csn.net>
//

static char rcsid[] = "$Id: xtvremote.cc,v 1.11 1997/03/01 06:09:49 smp Exp smp $";


#include <iostream.h>
#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <ctype.h>

#include <sys/ioctl.h>
#define TUNER
#include <machine/ioctl_meteor.h>

// Xlib headers
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/Xos.h>

// Standard Xt headers
#include <X11/Intrinsic.h>
#include <X11/StringDefs.h>
#include <X11/Shell.h>

// FWF Widget Set headers
#include <Xfwf/Board.h>
#include <Xfwf/Label.h>
#include <Xfwf/Button.h>
#include <Xfwf/PullDown.h>
#include <Xfwf/OptButton.h>
#include <Xfwf/TextMenu.h>
#include <Xfwf/RowCol.h>


class Action {

 public:

    Action() {}
    virtual ~Action() {}
    virtual int action() { return -1; }
};

class System : public Action {

 public:

    System( const char* args );
    virtual ~System() { delete _args; }
    virtual int action( void );

 private:

    char*	_args;

};

typedef		int(*func_t)(void*);

class Function : public Action {

 public:

    Function( int (*function)(char*), const char* args );
    Function( int (*function)(void) );
    virtual ~Function() { if ( _args ) delete _args; }
    virtual int action( void );

// private:

    int		(*_func)(void*);
    char*	_args;

};

#define USE_ACCELERATORS_NOT

#define OK			0
#define ERROR			-1

#define SKIPWS(P)		while( isspace((*P)) ) ++P

// path for device specific resource files
#define RES_BASE		BASE "/app-defaults/"

#define BUTTON_RES_TMPL		"%s_b%02d"
#define MESSAGE_RES_TMPL	"%s_m%02d"

// maximum number of buttons per device
#define NUM_BUTTONS		100

// maximum number of devices
#define NUM_DEVICES		16
#define DEFAULT_DEVICE		"1"

// instance & class names for X resource access
const static char* appName = "xtvremote";
const static char* appClass = "Xtvremote";

// 
char* displayFontName = "*-helvetica-bold-r-*-140-*";


//
// application resource setup
//
#define	XtNxDevice	"xDevice"
#define	XtCXDevice	"XDevice"
#define	XtNxDeviceList	"xDeviceList"
#define	XtCXDeviceList	"XDeviceList"
#define	XtNxFavorites	"xFavorites"
#define	XtCXFavorites	"XFavorites"
#define	XtNxDisplayFont	"xDisplayFont"
#define	XtCXDisplayFont	"xDisplayFont"

typedef struct {
    char*	xDevice;
    char*	xDeviceList;
    char*	xFavorites;
    char*	xDisplayFont;
} appResources_t, *appResourcesPtr;

static appResources_t	appResources;

static XtResource	resources[] = {
#define offset(field)		XtOffset(appResourcesPtr, field)

	{ XtNxDevice, XtCXDevice, XtRString, sizeof(char*),
		offset(xDevice), XtRImmediate, (XtPointer)DEFAULT_DEVICE },

	{ XtNxDeviceList, XtCXDeviceList, XtRString, sizeof(char*),
		offset(xDeviceList), XtRImmediate, (XtPointer)NULL },

	{ XtNxFavorites, XtCXFavorites, XtRString, sizeof(char*),
		offset(xFavorites), XtRImmediate, (XtPointer)NULL },

	{ XtNxDisplayFont, XtCXDisplayFont, XtRString, sizeof(char*),
		offset(xDisplayFont), XtRImmediate, (XtPointer)NULL },
#undef offset
};

static XrmOptionDescRec options[] = {
	{ "-d",	".xDevice",	XrmoptionSepArg, (caddr_t) DEFAULT_DEVICE },
	{ "-l",	".xDeviceList",	XrmoptionSepArg, (caddr_t) NULL },
	{ "-l",	".xFavorites",	XrmoptionSepArg, (caddr_t) NULL },
	{ "-l",	".xDisplayFont",	XrmoptionSepArg, (caddr_t) NULL },
};


//
// global variables
//
static int	xDevice = 0;

//
// data for pulldown menu
//
typedef struct {
    int		index;		// index in array
    char*	label;		// ascii label
    char*	resFile;	// resource file
} deviceTag;


// label to x device associations for device menu
deviceTag deviceList[ NUM_DEVICES + 1 ] = {
    { -1, NULL,	NULL }
};

typedef struct {
    WidgetList	w;		// an array of Widget
    Action**	a;		// an array of Action
    int		c;		// count of each array
    int		x;		// width
    int		y;		// heigth
} device;

device		devices[ NUM_DEVICES ];

#define MAX_CHANNEL	125
static u_char	favoriteChannels[ MAX_CHANNEL ];
static int	favoriteCount = 0;



// button data
typedef struct {
    Widget	w;
    Action*	a;
} button;

button		buttons[ NUM_BUTTONS ];

Widget		topLevel;

// connection stuff
static Action*	addMessage( char* args );

// callback functions
static void optionsCB( Widget, XtPointer, XtPointer );
static void deviceCB( Widget, XtPointer, XtPointer );
static void buttonCB( Widget w, XtPointer clientData, XtPointer callData );

// other local functions

static int	buildDeviceList( appResourcesPtr appRes, deviceTag* list );
static int	buildFavorites( appResourcesPtr appRes, u_char* list );
static int	loadDeviceResources( Widget w, deviceTag* list );
static int	buildDeviceMenu( Widget menu, deviceTag* list, int& index );
static void	setActiveMenu( Widget menu, deviceTag* list );

static int	buildButtonBoard( Widget ButtonBoard, deviceTag* deviceList );
static int	createButton( const char* deviceName, Widget parent,
			      int number, XtCallbackProc callBack );

// the hook to the STB card
static int	video;

// funk
static Widget*	widgetPointer = NULL;
static Widget	enterInputWidget = NULL;
static Widget	displayChannelWidget = NULL;

static int	getChannel( void );

//
// constructor, store args for action
//
System::System( const char* args )
{
    // keep a copy of the args
    _args = new char[ strlen( args ) + 1 ];
    strcpy( _args, args );

    return;
}


//
// do local system() call
//
int
System::action( void )
{

    if ( _args )
    {
	system( _args );
    }

    return OK;
}


//
// constructor, store args for action
//
Function::Function( int (*function)(char*), const char* args )
{
    // keep a copy of the args
    _args = new char[ strlen( args ) + 1 ];
    strcpy( _args, args );

    _func = (func_t)function;

    return;
}


//
// constructor, store args for action
//
Function::Function( int (*function)(void) )
{
    _args = NULL;

    _func = (func_t)function;

    return;
}


//
// do local Function() call
//
int
Function::action( void )
{
    return _func( _args );
}


//
// 
//
int
main( int argc, char **argv )
{
    XtAppContext	appContext;
    Widget		mainBox;
    Widget		mainMenu;
    Widget		optionsMenu;
    Widget		deviceMenu;
    Widget		ButtonBoard;

    // create top level shell
    topLevel = XtVaAppInitialize( &appContext, appClass,
				  options, XtNumber(options),
				  &argc, argv, NULL, NULL );

    // process remaining commandline args
    XtGetApplicationResources( topLevel,
			       &appResources, resources, XtNumber(resources),
			       0, 0 );

    if ((video = open("/dev/bktr0", O_RDONLY)) < 0) {
        fprintf(stderr, "open failed: %s\n", strerror(errno));
	exit( 1 );
    }


    // desired default device
    xDevice = atoi( appResources.xDevice ) - 1;

    if ( buildDeviceList( &appResources, deviceList ) == ERROR )
    {
	cerr << "BAD device list, missing Xremote app-default?" << endl;
	exit( 1 );
    }

    // 
    if ( buildFavorites( &appResources, favoriteChannels ) == ERROR )
    {
	cerr << "BAD favorites list, missing Xremote app-default?" << endl;
	exit( 1 );
    }

    // load device specific rdbs
    loadDeviceResources( topLevel, deviceList );

    // the top level container
    mainBox = XtVaCreateManagedWidget( "mainBox", xfwfBoardWidgetClass,
				       topLevel, NULL );

    //
    // build the menu bar
    //
    mainMenu = XtVaCreateManagedWidget( "menu", xfwfBoardWidgetClass,
					mainBox, NULL );

    optionsMenu = XtVaCreateManagedWidget( "oMenu", xfwfPullDownWidgetClass,
					   mainMenu, NULL );

    XtAddCallback( optionsMenu, XtNactivate, optionsCB, NULL );

    // create the device menu
    deviceMenu = XtVaCreateManagedWidget( "dMenu", xfwfOptionButtonWidgetClass,
					  mainMenu, NULL );

    XtAddCallback( deviceMenu, XtNactivate, deviceCB, NULL);

    // build device menu
    if ( buildDeviceMenu( deviceMenu, deviceList, xDevice ) == ERROR )
    {
	cerr << "BAD menu data" << endl;
	exit( 1 );
    }

#if defined( USE_ACCELERATORS )
    XtInstallAllAccelerators( topLevel, topLevel );
    XtInstallAllAccelerators( optionsMenu, topLevel );
#endif  // USE_ACCELERATORS

    //
    // button container
    //
    ButtonBoard = XtVaCreateManagedWidget( "keys", xfwfBoardWidgetClass,
					   mainBox, NULL );

    // create the button arrays
    cerr << "Building buttons, be patient, this takes time...";
    buildButtonBoard( ButtonBoard, deviceList );
    cerr << "done!" << endl;

    // activate only those devices that built correctly
    setActiveMenu( deviceMenu, deviceList );

    // change size
    Arg args[ 2 ];
    int i = 0;
    XtSetArg( args[i], XtNwidth, devices[ xDevice ].x );  i++;
    XtSetArg( args[i], XtNheight, devices[ xDevice ].y );  i++;
    XtSetValues( topLevel, args, i );

    // load a larger font for the channel display
    if ( displayChannelWidget )
    {
        XFontStruct* fontInfo = NULL;
        Display* display = XtDisplay( topLevel );

        if ( appResources.xDisplayFont )
        {
            fontInfo = XLoadQueryFont( display, appResources.xDisplayFont );
            if ( fontInfo == NULL )
            {
                fprintf( stderr, "couldn't load font: '%s'\n",
                         appResources.xDisplayFont );
            }
        }

        if ( fontInfo == NULL )
        {
            fontInfo = XLoadQueryFont( display, displayFontName );
            if ( fontInfo == NULL )
            {
                fprintf( stderr, "couldn't load font: '%s'\n",
                         displayFontName );
            }
        }

        if ( fontInfo )
        {
            XtVaSetValues( displayChannelWidget, XtNfont, fontInfo, NULL );
        }
    }

    // attempt to get the current channel selection
    getChannel();

    // 
    XtManageChildren( devices[ xDevice ].w, devices[ xDevice ].c );

    // make it all happen
    XtRealizeWidget( topLevel );

    // ...if only this was done with threads...
    XtAppMainLoop( appContext );
}


//
// 
//
static int
buildDeviceList( appResourcesPtr appRes, deviceTag* list )
{
    int index = 0;

    if ( appRes->xDeviceList == NULL ) return ERROR;

    // get string into a local buffer
    char* s = new char[ strlen(appRes->xDeviceList) + 1 ];
    strcpy( s, appRes->xDeviceList );

    // get the first label
    char* p = strtok( s, "," );
    SKIPWS( p );

    // process each of count possible values
    for ( int x = 0; x < NUM_DEVICES; ++x )
    {
	// set the index
	list[ x ].index = x;

	// assign the label
	list[ x ].label = p;

	// parse the resource file name
	if ( (p = strtok( NULL, "," )) == NULL )
	{
	    cerr << "buildDeviceList(), missing resource file" << endl;
	    return ERROR;
	}
	SKIPWS( p );

	// check for absolute or relative path
	if ( (*p == '/')
	     || ((p[0] == '.') && (p[1] == '/'))
	     || ((p[0] == '.') && (p[1] == '.') && (p[2] == '/')) )
	{
	    list[ x ].resFile = new char[ strlen( p ) + 1 ];
	    strcpy( list[ x ].resFile, p );
	}

	// no path, use default
	else
	{
	    list[ x ].resFile = new char[ strlen( p )
					  + strlen( RES_BASE ) + 1 ];
	    strcpy( list[ x ].resFile, RES_BASE );
	    strcat( list[ x ].resFile, p );
	}

	// get the next label
	if ( (p = strtok( NULL, "," )) == NULL ) break;
	SKIPWS( p );
    }

    // NULLify rest of list
    for ( ++x; x < NUM_DEVICES; ++x )
    {
	list[ x ].index = -1;
	list[ x ].label = NULL;
	list[ x ].resFile = NULL;
    }

    return OK;
}


//
// 
//
static int
buildFavorites( appResourcesPtr appRes, u_char* list )
{
    int index = 0;

    if ( appRes->xFavorites == NULL ) return ERROR;

    // get string into a local buffer
    char* s = new char[ strlen(appRes->xFavorites) + 1 ];
    strcpy( s, appRes->xFavorites );

    // get the first channel
    char* p = strtok( s, "," );
    SKIPWS( p );

    // process each of count possible values
    for ( int x = 0; x < MAX_CHANNEL; ++x )
    {
	// get the channel
	list[ x ] = atoi( p );

	// get the next channel
	if ( (p = strtok( NULL, "," )) == NULL ) break;
	SKIPWS( p );
    }

    // NULLify rest of list
    for ( favoriteCount = ++x; x < NUM_DEVICES; ++x )
    {
	list[ x ] = 0;
    }

    return OK;
}


//
// 
//
static int
loadDeviceResources( Widget w, deviceTag* list )
{
    XrmDatabase rdb = XtScreenDatabase( XtScreen( w ) );

    for ( int x = 0; x < NUM_DEVICES; ++x )
    {
	if ( list[ x ].label == NULL ) break;

	// attempt to load rdb for device
	const char* s = list[ x ].resFile;
	if ( XrmCombineFileDatabase( s, &rdb, True /**False*/ ) == 0 )
	{
	    list[ x ].index = -1;
	    list[ x ].resFile = NULL;
	    cerr << "resFile NOT found: " << s << endl;
	}
    }

    return OK;
}


//
// 
//
static int
buildDeviceMenu( Widget menu, deviceTag* list, int& index )
{
    int x, c;
    Arg args[ 10 ];

    // compute total length
    for ( c = 0, x = 0; x < NUM_DEVICES; ++x )
    {
	if ( list[ x ].label == NULL ) break;
	c += (strlen( list[ x ].label ) + 2);
    }
    ++c;

    // build menu
    char* s = new char[ c ];
    s[ 0 ] = '\0';
    for ( x = 0; x < NUM_DEVICES; ++x )
    {
	if ( list[ x ].label == NULL ) break;	// end of list
	strcat( s, list[ x ].label );
	strcat( s, "\n" );
    }

    // find a valid default entry
    if ( index == -1 ) index = 0;
    if ( list[ index ].index == -1 )
    {
	for ( x = 0; x < NUM_DEVICES; ++x )
	{
	    if ( list[ x ].index != -1 )
	    {
		index = x;
		break;
	    }
	}
	if ( x == NUM_DEVICES ) return ERROR;
    }

    // set the menu string and default label
    int i = 0;
    XtSetArg( args[i], XtNlabel, list[ index ].label );  i++;
    XtSetArg( args[i], XtNmenu, s );  i++;
    XtSetValues( menu, args, i);

    return OK;
}


//
// 
//
static void
setActiveMenu( Widget menu, deviceTag* list )
{
    // search list for (unNULLified) entries
    long active = 0x00000000;
    for ( int x = 0; x < NUM_DEVICES; ++x )
    {
	if ( list[ x ].index != -1 )		// valid entry
	{
	    active |= (1 << x);			// set bit in mask
	}
    }

    // find menu widget [created automatically by PullDown]
    Widget pw;
    Arg args[ 2 ];
    XtSetArg( args[ 0 ], XtNpopup, &pw );
    XtGetValues( menu, args, 1 );

    // set the 'active button list'
    XtSetArg( args[ 0 ], XtNactive, active );
    XtSetValues( pw, args, 1 );
}


static int
buildButtonBoard( Widget ButtonBoard, deviceTag* list )
{
    int count;
    int last;

    for ( int x = 0; x < NUM_DEVICES; ++x )
    {
	// end of list
	if ( list[ x ].label == NULL ) break;

	// handle sparse lists
	if ( list[ x ].index == -1 )
	{
	    devices[ x ].w = NULL;
	    devices[ x ].a = NULL;
	    continue;
	}
	/// cerr << "device: '" << list[ x ].label << "'" << endl;

	// clear the temp array
	for ( int xx = 0; xx < NUM_BUTTONS; ++xx )
	{
	    buttons[ xx ].w = NULL;
	    buttons[ xx ].a = NULL;
	}
	count = 0;
	last = -1;

	// create individual buttons
	const char* label = strrchr( list[ x ].resFile, '/' );
	if ( label == NULL ) label = list[ x ].resFile;
	else ++label;

	// build each defined button
	for ( xx = 0; xx < NUM_BUTTONS; ++xx )
	{
	    switch ( createButton( label,
				   ButtonBoard, xx, buttonCB ) )
	    {
	    case TRUE:
		++count;
		break;

	    case -1:
		last = xx;
		xx = NUM_BUTTONS;
		break;

	    case FALSE:
		break;
	    }
	}
	if ( last == -1 ) last = NUM_BUTTONS;

	// record count
	if ( (devices[ x ].c = count) == 0 )
	{
	    devices[ x ].w = NULL;
	    devices[ x ].a = NULL;
	    list[ x ].index = -1;
	    continue;
	}

	// build dense array of widgets
	int yy;
	devices[ x ].w = new Widget[ count ];
	for ( yy = 0, xx = 0; xx < last; ++xx )
	{
	    if ( buttons[ xx ].w )
	    {
		devices[ x ].w[ yy++ ] = buttons[ xx ].w;
	    }
	}

	if ( yy != count )
	{
	    cerr << "buildButtonBoard(), w != count" << endl;
	    cerr << "w: " << yy << "count: " << count << endl;
	    exit( 1 );
	}

	// build dense array of objects
	devices[ x ].a = new Action*[ count ];
	for ( yy = 0, xx = 0; xx < last; ++xx )
	{
	    if ( buttons[ xx ].a )
	    {
		// setup callback
		XtAddCallback( buttons[ xx ].w, XtNactivate,
			       buttonCB, (XtPointer)yy );
		// pack the object
		devices[ x ].a[ yy++ ] = buttons[ xx ].a;
	    }
	}

	if ( yy != count )
	{
	    cerr << "buildButtonBoard(), a != count" << endl;
	    cerr << "a: " << yy << "count: " << count << endl;
	    exit( 1 );
	}

	//
	char name[ 80 ];
	sprintf( name, "%s_size", label );

	// create an instance of it
	Widget w;
	w = XtVaCreateWidget( name, xfwfLabelWidgetClass, ButtonBoard, NULL );

	// get its label
	String message;
	Arg args[ 1 ];
	XtSetArg( args[ 0 ], XtNlabel, &message );
	XtGetValues( w, args, 1 );

	// delete label widget
	XtDestroyWidget( w );
	sscanf( message, "%dx%d", &devices[ x ].x, &devices[ x ].y );
	XtFree( message );
	/// cerr << "x: " << devices[x].x << ", y: " << devices[x].y << endl;
    }

    return OK;
}


//
// create push buttons
//
static int
createButton( const char* deviceName, Widget parent,
	       int number, XtCallbackProc callBack )
{
    // create a name for the message
    char name[ 256 ];
    sprintf( name, MESSAGE_RES_TMPL, deviceName, number );

#if defined( USE_XRMGETRESOURCE )

    would this be better???

#else  // USE_XRMGETRESOURCE

    // create an instance of it
    Widget w;
    w = XtVaCreateWidget( name, xfwfLabelWidgetClass, parent, NULL );

    // get its label
    String message;
    Arg args[ 1 ];
    XtSetArg( args[ 0 ], XtNlabel, &message );
    XtGetValues( w, args, 1 );

    // delete label widget
    XtDestroyWidget( w );

#endif // USE_XRMGETRESOURCE

    // test it
    if ( message )
    {
	Action* action = addMessage( message );
	XtFree( message );

	// found last message
	if ( action == NULL ) return (-1);

	// record buttons action
	buttons[ number ].a = action;

	// create the actual button
	sprintf( name, BUTTON_RES_TMPL, deviceName, number );
	w = XtVaCreateWidget( name, xfwfButtonWidgetClass, parent, NULL );

        // funky...
        if ( widgetPointer )
        {
            *widgetPointer = w;
            widgetPointer = NULL;
        }

	// record widget
	buttons[ number ].w = w;
	return TRUE;
    }

    else
    {
	buttons[ number ].w = NULL;
	buttons[ number ].a = NULL;
	return FALSE;
    }
}


//
// options menu callback
//
static void
optionsCB( Widget w, XtPointer clientData, XtPointer callData )
{
    XfwfTextMenuData& data = *(XfwfTextMenuData*)callData;

    switch ( data.n )
    {
    case 0:
	exit(0);

    default:
	break;
    }
}


//
// device menu callback
//
static void
deviceCB( Widget w, XtPointer clientData, XtPointer callData )
{
    XfwfTextMenuData& data = *(XfwfTextMenuData*)callData;

    // check for invalid device
    if ( deviceList[ data.n ].index == -1 )
    {
	return;
    }

    // unmap the current device buttons
    XtUnmanageChildren( devices[ xDevice ].w, devices[ xDevice ].c );

    // get the new device type
    xDevice = deviceList[ data.n ].index;

    // change size
    Arg args[ 2 ];
    int i = 0;
    XtSetArg( args[i], XtNwidth, devices[ xDevice ].x );  i++;
    XtSetArg( args[i], XtNheight, devices[ xDevice ].y );  i++;
    XtSetValues( topLevel, args, i );

    // map the new device buttons
    XtManageChildren( devices[ xDevice ].w, devices[ xDevice ].c );
}


//
// button pressed callback
//
static void
buttonCB( Widget w, XtPointer clientData, XtPointer callData )
{
    if ( devices[ xDevice ].a[ (int)clientData ] )
	devices[ xDevice ].a[ (int)clientData ]->action();
}


/////////////////////////////// tuner functions ///////////////////////////////

static int	currentChannel = 0;
static int	previousChannel = 0;
static int	inputChannel = 0;
static int	inputDigits = 0;


//
// 
//
static int
updateChannel( int channel )
{
    char label[ 128 ];

    previousChannel = currentChannel;
    currentChannel = channel;

#if 0
    XfwfLabelSetLabel( displayChannelWidget, label );
#else
    sprintf( label, "Channel %d", channel );
    XtVaSetValues( displayChannelWidget, XtNlabel, label, NULL );
#endif
}


//
// 
//
static int
setChannel( char* arg )
{
    int x;

    x = atoi( arg );
    if ( ioctl( video, TVTUNER_SETCHNL, &x ) < 0 )
    {
        fprintf( stderr, "ioctl failed: %s\n", strerror( errno ) );
        return ERROR;
    }

    updateChannel( x );

    return OK;
}


//
// 
//
static int
upChannel( void )
{
    int x;

    if ( currentChannel )
    {
        for ( int y = 0; y < favoriteCount; ++y )
        {
            if ( favoriteChannels[ y ] > currentChannel )	// found
            {
                x = y;
                break;
            }
        }

        if ( y == favoriteCount )				// wrapped
        {
            x = 0;
        }
    }
    else
    {
        x = 0;
    }

    if ( !favoriteChannels[ x ] ) return ERROR;

    x = favoriteChannels[ x ];
    if ( ioctl( video, TVTUNER_SETCHNL, &x ) < 0 )
    {
        fprintf( stderr, "ioctl failed: %s\n", strerror( errno ) );
        return ERROR;
    }

    updateChannel( x );

    return OK;
}


//
// 
//
static int
downChannel( void )
{
    int x;

    if ( currentChannel )
    {
        for ( int y = favoriteCount - 1; y > -1 ; --y )
        {
            if ( favoriteChannels[ y ] < currentChannel )	// found
            {
                x = y;
                break;
            }
        }

        if ( y == -1 )						// wrapped
        {
            x = favoriteCount - 1;
        }
    }
    else
    {
        x = favoriteCount - 1;
    }

    if ( !favoriteChannels[ x ] ) return ERROR;

    x = favoriteChannels[ x ];
    if ( ioctl( video, TVTUNER_SETCHNL, &x ) < 0 )
    {
        fprintf( stderr, "ioctl failed: %s\n", strerror( errno ) );
        return ERROR;
    }

    updateChannel( x );

    return OK;
}


//
// 
//
static int
getChannel( void )
{
    int x;

    if ( ioctl( video, TVTUNER_GETCHNL, &x ) < 0 )
    {
        fprintf( stderr, "ioctl failed: %s\n", strerror( errno ) );
        return ERROR;
    }

    updateChannel( x );

    return x;
}


//
// 
//
static int
getStatus( void )
{
    int x;

    if ( ioctl( video, TVTUNER_GETSTATUS, &x ) < 0 )
    {
        fprintf( stderr, "ioctl failed: %s\n", strerror( errno ) );
        return ERROR;
    }

    fprintf( stderr, "0x%02x\n", x );

    return x;
}


//
// 
//
static int
setPrevious( void )
{
    int x;

    if ( !previousChannel ) return 0;

    x = previousChannel;
    if ( ioctl( video, TVTUNER_SETCHNL, &x ) < 0 )
    {
        fprintf( stderr, "ioctl failed: %s\n", strerror( errno ) );
        return ERROR;
    }

    updateChannel( x );

    return x;
}


static void
enterInputLabel( void )
{
    char label[ 128 ];

#if 0
    XfwfLabelSetLabel( enterInputWidget, label );
#else
    if ( inputChannel )
        sprintf( label, "Enter\n%d", inputChannel );
    else
        strcpy( label, "Enter" );

    XtVaSetValues( enterInputWidget, XtNlabel, label, NULL );
#endif
}

//
// 
//
static int
enterInput( void )
{
    int x;

    if ( !inputChannel ) return 0;

    x = inputChannel;
    inputChannel = inputDigits = 0;

    if ( ioctl( video, TVTUNER_SETCHNL, &x ) < 0 )
    {
        fprintf( stderr, "ioctl failed: %s\n", strerror( errno ) );
        return ERROR;
    }

    updateChannel( x );

    return x;
}

//
// 
//
static int
addInput( char* input )
{
    if ( inputDigits >= 3 )
    {
        inputChannel = 0;
        inputDigits = 0;
    }

    ++inputDigits;
    inputChannel = (inputChannel * 10) + atoi( input );

    enterInputLabel();

    return OK;
}

//
// 
//
static int
clearInput( void )
{
    inputChannel = 0;

    enterInputLabel();

    return OK;
}

static int
noop( void )
{
}


//
// create an action object for the button
//
static Action*
addMessage( char* msg )
{
    Action*	object;
    char*	stringp;

    // make a working copy
    char* args = new char[ strlen( msg ) + 1 ];
    strcpy( args, msg );

    // break string up into address & message
    stringp = args;
    strsep( &stringp, " 	" );

    // check for end of list
    if ( strcmp( args, "MSG_END" ) == 0 )
    {
	return NULL;
    }

    // create a System action
    else if ( strcmp( args, "SYSTEM" ) == 0 )
    {
	object = new System( stringp );
    }

    // select a channel
    else if ( strcmp( args, "SET_CHNL" ) == 0 )
    {
	object = new Function( setChannel, stringp );
    }

    // get the current channel
    else if ( strcmp( args, "GET_CHNL" ) == 0 )
    {
	object = new Function( getChannel );
    }

    // get tuner internal status
    else if ( strcmp( args, "GET_STATUS" ) == 0 )
    {
	object = new Function( getStatus );
    }

    // select the previously selected channel
    else if ( strcmp( args, "PREVIOUS" ) == 0 )
    {
	object = new Function( setPrevious );
    }

    // create a channel 'Enter' object
    else if ( strcmp( args, "ENTER_INPUT" ) == 0 )
    {
	object = new Function( enterInput );
        widgetPointer = &enterInputWidget;
    }

    // create a button to add input to the 'Enter' object
    else if ( strcmp( args, "ADD_INPUT" ) == 0 )
    {
	object = new Function( addInput, stringp );
    }

    // clear the current contents of the 'Enter' object
    else if ( strcmp( args, "CLEAR_INPUT" ) == 0 )
    {
	object = new Function( clearInput );
    }

    // create a display for the current channel
    else if ( strcmp( args, "DISPLAY_CHNL" ) == 0 )
    {
	object = new Function( noop );
        widgetPointer = &displayChannelWidget;
    }

    // create a ? action
    else if ( strcmp( args, "UP_CHNL" ) == 0 )
    {
	object = new Function( upChannel );
    }

    // create a ? action
    else if ( strcmp( args, "DN_CHNL" ) == 0 )
    {
	object = new Function( downChannel );
    }

    // create a no-op action
    else if ( strcmp( args, "NOOP" ) == 0 )
    {
	object = new Function( noop );
    }

    // 
    else
    {
	object = new Function( noop );
    }

    // return memory
    delete args;

    return object;
}
